# Algorithm for dominance matrix generation 
# inputs: names of objects
#         X=numpy.array() array of values 


import math
import copy
import random
import Tkinter
import rmod2
from ScrolledText import *
import Canvas
import tkFileDialog
import raioop2

# 30.3.2013 A renewed trial to adapt Wielands fuzzy program for PyHasse
# 30.3.2013 Umsellung scheint gelungen. Dom und Sep-Matrizen sind plausibel


# some global constants 
EQ=0
GT=1
LT=-1
NC=2
OK=0
NN=1
DELTA=0.0  # to distinguage between two values


class simple_sit():
    # Wieland, March, 2012
    def __init__(self,_name,_feld):
        self.name=_name
        self.feld=copy.copy(_feld)
        self.rand=copy.copy(_feld)
    def rand_uniform(self,p):  # p=percentage of change
        for i in range(len(self.rand)):
            self.rand[i]=self.feld[i]+\
                self.feld[i]*p/100.0*(0.5-random.random())
    
            
class compare_sit():
    # Wieland March 2012
    def __init__(self,sit_list,_eps,_d,*_p):  # d=distribution, p=parameter
        self.sits=sit_list
        self.ni=len(self.sits)
        self.nj=len(self.sits[0].feld) # number of qi
        self.dom=rmod2.marray(self.ni,self.ni)
        self.sep=rmod2.marray(self.ni,self.ni)
                
        self.d=_d    # distribution 0=uniform, 2=normal
        self.n=1
        if(self.d==0 or self.d==1):
            self.p=_p[0] # parameter percentage or sd
        if(self.d==2):   # triangular distribution (left,mode,right)
            self.p=_p[0]
            self.left=_p[1]
            self.mode=_p[2]
            self.right=_p[3]
        self.eps=_eps # epsilon

    
    def compare(self,feld1,feld2): # help function to compare two vals
        if(len(feld1)!=len(feld2)):
            return NC
        val,valk,valg=0,0,0
        for i in range(len(feld1)):
            if(math.fabs(feld1[i]-feld2[i])>DELTA):
                val=1
                if(feld1[i]<feld2[i]): 
                    valk=1
                if(feld1[i]>feld2[i]): 
                    valg=1;
        if(val==0):
            return EQ
        if(valk==1 and valg==1): 
            return NC  # not comparable
        if(valk==0 and valg==1): 
            return GT  # return GT
        if(valk==1 and valg==0):
            return LT  # return LT   

    def gen_next(self,i,n): # change only one sit i with distribution d
        self.n=n
        for k in range(n):
            if(self.d==0):
                self.sits[i].rand_uniform(self.p)
            
            for j in range(self.ni):
                if(i!=j):
                    c=self.compare(self.sits[i].rand,self.sits[j].feld)
                    if(c==GT):
                        self.dom[i][j]+=1
                        continue
                    if(c==LT):
                        self.dom[j][i]+=1
                        continue
                    if(c==NC):
                        self.sep[i][j]+=1
                        self.sep[j][i]+=1
                        continue
                    if(c==EQ):
                        if(random.random()>0.5):
                            self.dom[i][j]+=1
                        else:
                            self.dom[j][i]+=1
                        continue
    def gen_all(self,n): # make changes simultaneously
        #print "MC-simulations = ",n
        for k in range(n):
            for sit in self.sits:
                if(self.d==0):
                    sit.rand_uniform(self.p)
                    continue
                
            for i in range(self.ni):
                for j in range(self.ni): #korrigiert 19.4.2012
                    if(j!=i):
                        c=self.compare(self.sits[i].rand,self.sits[j].rand)
                        if(c==GT):
                            self.dom[i][j]+=1
                            continue
                        if(c==LT):
                            self.dom[j][i]+=1
                            continue
                        if(c==NC):
                            self.sep[i][j]+=1
                            self.sep[j][i]+=1  #brg 18.4.2012 wieder eingeschaltet
                            continue
                        if(c==EQ):
                            if(random.random()>0.5):
                                self.dom[i][j]+=1
                            else:
                                self.dom[j][i]+=1
                        continue



def fileselect():
    # Brg 31.3.2013
    global fn,dm,obj,prop,r,k
    fn=tkFileDialog.askopenfilename()
    dm,ob,pro,r,k=rmod2.EXCELreadin4(fn)
    lb1.config(text="file ="+str(fn)+"\nwith "+str(r)+"rows\n"+str(k)+"columns\n opened")
    prop=[]
    for j in range(0,k):
        prop.append(pro[j+1])
            
    obj=rmod2.vecgentext1(r)     #8.5.08 changed to ...text1
    for i in range(0,r):
        obj[i]=ob[i+1]


def HD():
    """based on objred (zeta-matrix) the Hasse diagram is generated and one linear extension)"""
    # Brg 20.12.07, 28.12.07, 2.10.,3.10.09
    # If the number of representants > 70 then only the cover-matrix and the level-population
    #
    global r,k,rred,objred,cov
    hfhd2=Tkinter.Toplevel(hf)
    hfhd2.title('Hasse diagram / cover matrix ')
    freq=Tkinter.Frame(hfhd2,width=200,height=300)
    freq.pack()
    
    rred,xred,objred,txt,neq=rmod2.equivcompactm(r,k,dm,obj)  
    
    #
    zetahilf=rmod2.comp(rred,k,xred)  #zeta matrix with col<row
    zeta=rmod2.marray(rred,rred)
    for i in range(0,rred):  #however we need for the 'normal' presentation: row < col. Transposition
        for j in range(0,rred):
            zeta[i][j]=zetahilf[j][i]
    #
    #One linear extension (based on levels)
    
    lobj,lext,revlext,zzmax=rmod2.levtopsort(rred,zeta)
       
    levmax=-10**20     #20.6.08
    for i in range(0,zzmax):
        if len(lobj[i])>=levmax:
            levmax=len(lobj[i])
       
    #Adj-Listen

    adjc=rmod2.adjacent(rred,lext,revlext,zeta)
        
    #Transitive reduced relations
    ts,tr=rmod2.trreduct(rred,adjc,lext)
       
    
    cov=rmod2.cover(rred,tr,lext)

    if rred>100:
        lbhd3=Tkinter.Label(freq,text=" +++too many objects ++++  ",bg="yellow")
        lbhd3.pack()
        cover_show()
    else:
        lbhd3=Tkinter.Label(freq,text=" Hasse Diagram  ")
        lbhd3.pack()
        actwidth,actheight,factorx,\
                factory,geom0x,geom0y,garrow1x,garrow1y,garrow2y=rmod2.geomHD(rred,levmax,zzmax)
        cv=Tkinter.Canvas(freq,width=actwidth,height=actheight,bg="white")
        
        lz=zzmax
        l=rmod2.vecgen1(lz)
        for i in range(0,lz):
            l[i]=len(lobj[i])

        pos=rmod2.vecgen1(lz)
        for i in range(0,lz):
            pos[i]=l[i]

        geomx=rmod2.marray(rred,rred)
        geomy=rmod2.marray(rred,rred)

       
        for i in range(0,lz):
            for j in range(0,l[i]):
                geomx[i][j]=geom0x+(factorx/(pos[i]+1.0))*(j+1)  #10.6.08 160-480
                geomy[i][j]=geom0y+(factory/lz)*i
            
        
        for i in range(0,lz-1):
            for j in range(0,l[i]):
                for i1 in range(i+1,lz):
                    for j1 in range(0,l[i1]):
                        if cov[lobj[i1][j1]][lobj[i][j]]==1:
                            cv.create_line(geomx[i][j],geomy[i][j],geomx[i1][j1],geomy[i1][j1])


        for i in range(0,lz):
            for j in range(0,l[i]):
                cv.create_oval(geomx[i][j]-10,geomy[i][j]+10,geomx[i][j]+10,geomy[i][j]-10,fill="white")
                cv.create_text(geomx[i][j],geomy[i][j],text=objred[lobj[i][j]])
        
        cv.create_line(garrow1x,garrow1y,garrow1x,garrow2y-12,arrow="first",width=3,fill="red")
        cv.create_text(garrow1x+8,garrow1y-6,text="high")
        cv.create_text(garrow1x+8,garrow2y-6,text="low")
        cv.pack()
        buttcov=Tkinter.Button(freq,text="Cover-relations ",width=25,command=cover_show)
        buttcov.pack()


def showmatrices():
    # brg 4.8.2012, 31.3.2013
    tp2=Tkinter.Toplevel(hf)
    tp2.title('showDM')
    tfile=ScrolledText(tp2,width=60,height=4)
    tfile.pack(expand=1,fill="both")
    showDM(k,r,obj,prop,dm,tfile)
    

def showDM(k,r,obj,prop,matrix,tfile):
    """ Routine : shows the data matrix """
    #Brg 13.1.08, 4.8.2012, 31.3.2013
    
    txt1="\n"
    txt1+="\n objects\n"
    for i in range(0,r):
        txt1+=" "+obj[i]+",\t"
    txt1+="\n"
    txt2="\n properties  \n"
    for j in range(0,k):
        txt2+=" "+prop[j]+",\t"
    txt2+="\n"
    txt3="\n data matrix  \n"
    txt3+="\t"
    for j in range(0,k):
        txt3+=prop[j]+"\t"
    txt3+="\n"    
    for i in range(0,r):
        txt3+=obj[i]+"\t"
        for j in range(0,k):
            txt3+=str(matrix[i][j])+"\t"
        txt3+="\n"
    txt3+="\n\n"
    txt=txt1+txt2+txt3
    tfile.insert("end",txt)

def cover_show():
    """Cover-matrix is displayed"""
    #brg 28.12.07
    #*28.12.07
    #akt28.12.07, 31.3.2013
    #note: -
    #source: antag
    global r,k,rred
    tpcover=Tkinter.Toplevel(hf)
    tpcover.title(' cover-matrix ')
    
    lbhd2=Tkinter.Label(tpcover,text=" Cover-Relations ")
    lbhd2.pack()
    txt3="\n cover-matrix  \n"
    for i in range(0,rred):
        txt3+="\n"+objred[i]+":\t"
        txt3+="is cover_less than "
        for j in range(0,rred):
            if i!=j:
                if cov[i][j]==1:
                    txt3+=" "+objred[j]+" ,"
                else:
                    pass
            else:
                pass
        txt3+="\n"    
    txt3+="\n"
    txt3+=" -- end of list --"
    tcov=ScrolledText(tpcover,width=50, height=4)
    tcov.pack()
    tcov.insert("end",txt3)



def calc():
    # brg 31.3.2013
    # based on Wieland, ZALF
    global DOM,SEP,fuzzyzeta,r,k
    fuzzyzeta=rmod2.marray(r,r)
    try:
        eps=int(e_eps.get())
    except:
        eps=250
    try:    
        MC=int(e_MC.get())
    except:
        MC=1000
    try:    
        percent=int(e_percent.get())
    except:
        percent=10
    
    fobj=0
    sit_list=[]
    for i in dm:
        #print obj,i
        fobj+=1
        sit_list.append(simple_sit(fobj,i))
    comp=compare_sit(sit_list,eps,0,percent)
    comp.gen_all(MC)
    for i in range(0,r):
        for i1 in range(0,r):
            if i==i1:
                fuzzyzeta[i][i1]=1
            if comp.dom[i][i1]>eps:
                fuzzyzeta[i][i1]=1
    rmod2.matout(r,r,fuzzyzeta)
    DOM=rmod2.marray(r,r)
    SEP=rmod2.marray(r,r)
    for i in range(0,r):
        for i1 in range(0,r):
            DOM[i][i1]=comp.dom[i][i1]
    for i in range(0,r):
        for i1 in range(0,r):
            SEP[i][i1]=comp.sep[i][i1]
    print "eps= ",eps,"  MC=  ",MC, "   percent= ",percent
    print "DOM= "
    rmod2.matout(r,r,DOM)
    print "SEP "
    rmod2.matout(r,r,SEP)


def extract():
    """extracts from the crisp matrix the order relations"""
    #brg 30.11.07, 11.4.08, 31.3.2013
    # the matrix order corresponds to objec
    # zeta is the matrix rs*rs which results after extract
    # rs is the number of rows/ columns of the square matrix order
    global zeta, objec,rs,r,covf
    try:
        eps=int(e_eps.get())
    except:
        eps=250.0
        
    eqtext="\n"
    eqtext+="eps : "+str(eps)+":  \n"
    
    
    ec=rmod2.dmcompress(r,r,fuzzyzeta)             #30.11. from cq the equivalent relations
    eqtext+=rmod2.eqcguiout1(r,ec,obj)        #30.11. matrix output
    tpequiv=Tkinter.Toplevel(hf)
    tpequiv.title(' eqivalence classes  ')
    teq=ScrolledText(tpequiv,width=70,height=3)
    teq.pack(expand="yes",fill="both")
    teq.insert("end",eqtext)
    txt=rmod2.irc_elim(r,ec)    #7.5.08 identifycolrows replaced by icr_elim
                                #in txt are the rows (and columns)
                                #...which can be eliminated
    ltxt=len(txt)
       
    b=rmod2.recogempty(txt,ltxt)
    # Not only the matrix cq is to be contracted but also the object vector obj
    # txt is a code of row numbers and as such ok for the matrixeliminations
    # however, to perform a new object list, txt must
    # be transformed in the corresponding object names
            
    an1,ldel1=rmod2.matrixcoldel(txt,r,r,fuzzyzeta) # columns eliminated
    zetahilf,ldel2=rmod2.matrixrowdel(txt,r,r-ldel1,an1)
    rs=r-ldel1
    zeta=rmod2.marray(rs,rs)
    for i in range(0,rs):
        for j in range(0,rs):
            zeta[i][j]=zetahilf[j][i]
    #

    
    if b=="true":
        #there is nothing to be reduced hence
        objec=rmod2.vecgentext1(r)
        for i in range(0,r):
            objec[i]=obj[i]
    else:        
        conobj=rmod2.vecgentext1(ltxt)
        for i in range(0,ltxt):
            conobj[i]=obj[int(txt[i])]
        #print "conobj = ",conobj    
        objec=rmod2.vecgentext1(rs)
        zaehl=0
        for i in range(0,r):
            #print "obj i  ",obj[i],"  in conobj  ",conobj
            if obj[i]not in conobj:
                objec[zaehl]=obj[i]
                zaehl+=1
        
    # HD based on zeta which is obtained from fuzzyzeta which is derived from DOM
    tpf=Tkinter.Toplevel(hf)
    tpf.title(' HD of fuzzyzeta ')
    lobjf,lextf,revlextf,zzmaxf=rmod2.levtopsort(rs,zeta)
       
    levmaxf=-10**20     #20.6.08
    for i in range(0,zzmaxf):
        if len(lobjf[i])>=levmaxf:
            levmaxf=len(lobjf[i])
       
    #Adj-Listen

    adjcf=rmod2.adjacent(rs,lextf,revlextf,zeta)
        
    #Transitive reduced relations
    ts,tr=rmod2.trreduct(rs,adjcf,lextf)
       
    
    covf=rmod2.cover(rs,tr,lextf)

    if rs>100:
        coverf_show()
    else:
        actwidth,actheight,factorx,\
                factory,geom0x,geom0y,garrow1x,garrow1y,garrow2y=rmod2.geomHD(rs,levmaxf,zzmaxf)
        cvf=Tkinter.Canvas(tpf,width=actwidth,height=actheight,bg="white")
        
        lz=zzmaxf
        l=rmod2.vecgen1(lz)
        for i in range(0,lz):
            l[i]=len(lobjf[i])

        pos=rmod2.vecgen1(lz)
        for i in range(0,lz):
            pos[i]=l[i]

        geomx=rmod2.marray(rs,rs)
        geomy=rmod2.marray(rs,rs)

       
        for i in range(0,lz):
            for j in range(0,l[i]):
                geomx[i][j]=geom0x+(factorx/(pos[i]+1.0))*(j+1)  #10.6.08 160-480
                geomy[i][j]=geom0y+(factory/lz)*i
            
        
        for i in range(0,lz-1):
            for j in range(0,l[i]):
                for i1 in range(i+1,lz):
                    for j1 in range(0,l[i1]):
                        if covf[lobjf[i1][j1]][lobjf[i][j]]==1:
                            cvf.create_line(geomx[i][j],geomy[i][j],geomx[i1][j1],geomy[i1][j1])


        for i in range(0,lz):
            for j in range(0,l[i]):
                cvf.create_oval(geomx[i][j]-10,geomy[i][j]+10,geomx[i][j]+10,geomy[i][j]-10,fill="white")
                cvf.create_text(geomx[i][j],geomy[i][j],text=objec[lobjf[i][j]])
        
        cvf.create_line(garrow1x,garrow1y,garrow1x,garrow2y-12,arrow="first",width=3,fill="red")
        cvf.create_text(garrow1x+8,garrow1y-6,text="high")
        cvf.create_text(garrow1x+8,garrow2y-6,text="low")
        cvf.pack()
        buttcovf=Tkinter.Button(tpf,text="Cover-relations ",width=25,command=coverf_show)
        buttcovf.pack()


def coverf_show():
    # Brg 31.3.2013
    """Cover-matrix is displayed"""
    #brg 28.12.07
    #*28.12.07
    #akt28.12.07, 31.3.2013
    #note: -
    #source: antag
    global r,k,rs
    tpcoverf=Tkinter.Toplevel(hf)
    tpcoverf.title(' cover_fuzzy-matrix ')
     
    txt3="\n cover-fuzzy-matrix  \n"
    for i in range(0,rs):
        txt3+="\n"+objec[i]+":\t"
        txt3+="is covered by "
        for j in range(0,rs):
            if i!=j:
                if covf[i][j]==1:
                    txt3+=" "+objec[j]+" ,"
                else:
                    pass
            else:
                pass
        txt3+="\n"    
    txt3+="\n"
    txt3+=" -- end of list --"
    tcovf=ScrolledText(tpcoverf,width=50, height=4)
    tcovf.pack()
    tcovf.insert("end",txt3)
    


def simpl_eval():
    # brg 31.3.2013
    # A HD of fuzzyzeta as outcome of fuzzy DDS is ok
    #...albeit errors may occur, when the fuzzyzeta-based directed
    #...graph has cyclic structures
    # Before I have a better idea I just give out DOM and SEP and
    # allow individual control of any two objects.
    global e_obj1,e_obj2
    tpsimpl=Tkinter.Toplevel(hf)
    tpsimpl.title(' Simple Evaluierung ')
    buttmatrixDOM=Tkinter.Button(tpsimpl,text=" Show the DOM-matrix ",width=25,command=show_DOM)
    buttmatrixDOM.pack()
    buttmatrixSEP=Tkinter.Button(tpsimpl,text=" Show the SEP-matrix ",width=25,command=show_SEP)
    buttmatrixSEP.pack()
    lbsimpl=Tkinter.Label(tpsimpl,text="Enter two objects of interest ")
    lbsimpl.pack()
    frobj=Tkinter.Frame(tpsimpl,bd=2,relief="sunken")
    frobj.pack()
    e_obj1=Tkinter.Entry(frobj,width=10)
    e_obj1.pack(side="left")
    e_obj2=Tkinter.Entry(frobj,width=10)
    e_obj2.pack(side="left")
    buttindivobj=Tkinter.Button(tpsimpl,text=" Information: two objects ",width=25,command=indivobj)
    buttindivobj.pack()

def show_DOM():
    # Brg 31.3.2013
    tpdom=Tkinter.Toplevel(hf)
    app1=raioop2.brgmatrixoutput3(tpdom,DOM,"DOM-matrix"," ",r,r,obj,obj," Dominances ")
    app1.ausgabe()

def show_SEP():
    # Brg 31.3.2013
    tpsep=Tkinter.Toplevel(hf)
    app2=raioop2.brgmatrixoutput3(tpsep,SEP,"SEP-matrix"," ",r,r,obj,obj," Separabilities ")
    app2.ausgabe()
    

def indivobj():
    # Brg 31.3.2013
    global r
    try:
        eps=int(e_eps.get())
    except:
        eps=250
        
    try:
        MC=int(e_MC.get())
    except:
        MC=1000

    try:
        percent=int(e_percent.get())
    except:
        percent=10
        
    obj1=e_obj1.get()
    obj2=e_obj2.get()
    iob1=obj.index(obj1)
    iob2=obj.index(obj2)
    vec1=rmod2.vecgen1(k)
    vec2=rmod2.vecgen1(k)
    txt="Basic information "+"\n"
    txt+="Selected file  = "+str(fn)+"\n"
    txt+="eps            = "+str(eps)+"\n"
    txt+="MC             = "+str(MC)+"\n"
    txt+="Noise-amplitude= "+str(percent)+"\n"
    
    txthilf1=""
    for j in range(0,k):
        vec1[j]=dm[iob1][j]
        txthilf1+=str(round(dm[iob1][j],3))+", "
    txthilf2=""                  
    for j in range(0,k):
        vec2[j]=dm[iob2][j]
        txthilf2+=str(round(dm[iob2][j],3))+", "              
    txt+="Dataprofiles"+"\n"
    txt+=str(obj1)+": "+txthilf1+"\n"
    txt+=str(obj2)+": "+txthilf2+"\n"
    txt+="\n"
    nmc=2.0*MC
    dom12=1.0*DOM[iob1][iob2]/nmc
    dom21=1.0*DOM[iob2][iob1]/nmc
    sep12=1.0*SEP[iob1][iob2]/nmc
    deltadom=dom12-dom21
    
    
    txt+="DOM and SEP of "+str(obj1)+"  and "+str(obj2)+"\n"
    txt+=60*"="+"\n"
    txt+="Dom("+str(obj1)+", "+str(obj2)+"):= "+str(round(dom12,2))+"\n"
    txt+="Dom("+str(obj2)+", "+str(obj1)+"):= "+str(round(dom21,2))+"\n"
    txt+="Sep("+str(obj1)+", "+str(obj2)+"):= "+str(round(sep12,2))+"\n"
    sumdomdomsep=dom12+dom21+sep12
    txt+="Sum of DOM(1,2)+Dom(2,1)+Sep(1,2) = "+str(sumdomdomsep)+"\n"
    txt+="Difference : "+"\n"
    txt+="Dom("+str(obj1)+", "+str(obj2)+") - Dom("+str(obj2)+", "+str(obj1)+")= " +str(round(deltadom,2))+"\n"
    tpindiv=Tkinter.Toplevel(hf)
    tindiv=ScrolledText(tpindiv,width=80,height=10)
    tindiv.pack()
    tindiv.insert("end",txt)
    






hf=Tkinter.Tk()
hf.title(' fuzzydds ')
aboutHDsimpl=raioop2.brgabout(hf,"tp1","about_fuzzydds.txt",7,\
            "March, 31th,2013"," about fuzzydds ", "white", "t1",120,5,ScrolledText)
helpHDsimpl=raioop2.brgabout(hf,"tp1","help_fuzzydds.txt",7,\
            "March, 31th,2013"," help_fuzzydds ", "white", "t1",120,5,ScrolledText)
w=25
col="cyan"
fr=Tkinter.Frame(hf,width=60)
fr.pack()
buttfile=Tkinter.Button(fr,text=" Select and open file ",bg=col,width=w,command=fileselect)
buttfile.pack()
lb1=Tkinter.Label(fr,text="")
lb1.pack()
buttshowfile=Tkinter.Button(fr,text=" Show data file ",width=w,bg=col,command=showmatrices)
buttshowfile.pack()
buttHD=Tkinter.Button(fr,text=" Show HD ",width=w,bg=col,command=HD)
buttHD.pack()
lb2=Tkinter.Label(fr,text="Entryfields for: eps (250), MC(1000), percent (10) ")
lb2.pack(pady=3)
frentry=Tkinter.Frame(fr,bd=2,relief="sunken")
frentry.pack()
e_eps=Tkinter.Entry(frentry,width=5)
e_eps.pack(side="left")
e_MC=Tkinter.Entry(frentry,width=5)
e_MC.pack(side="left")
e_percent=Tkinter.Entry(frentry,width=5)
e_percent.pack(side="left")
buttcalc=Tkinter.Button(fr,text="MonteCarlo - noise: Dominance and Separability matrices ",bg=col,width=w,command=calc)
buttcalc.pack()
buttextract=Tkinter.Button(fr,text="Based on Dom>eps : HD  ",bg=col,width=w,command=extract)
buttextract.pack()
buttsimpleval=Tkinter.Button(fr,text="Menue for \nfurther evaluation ",bg=col,width=w,command=simpl_eval)
buttsimpleval.pack()


buttexit=Tkinter.Button(fr,text="Exit ",bg="red",bd=4,width=w,command=hf.destroy)
buttexit.pack(pady=5)














hf.mainloop()
